/*
    Valtra Tractors
    Copyright 2004 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <math.h>
#include <stdlib.h>
#include "effects.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

#define ROTATEARGB

SDL_Surface *Effects::multAlpha(SDL_Surface *in, int alpha) {
	SDL_Surface *out = SDL_CreateRGBSurface(SDL_SWSURFACE,in->w,in->h,32,in->format->Rmask,in->format->Gmask,in->format->Bmask,in->format->Amask);
	if (SDL_MUSTLOCK(in)) SDL_LockSurface(in);
	if (SDL_MUSTLOCK(out)) SDL_LockSurface(out);

	Uint32 *inptr = (Uint32*)in->pixels;
	Uint32 *outptr = (Uint32*)out->pixels;

	int inw = in->w;
	int inh = in->h;
	int Amask = in->format->Amask;
	int Ashift = in->format->Ashift;

	for (int y = 0; y < inh; y++) {
		for (int x = 0; x < inw; x++) {
			Uint32 pixel = *inptr++;
			int tmp = (pixel&Amask)>>Ashift;
			tmp = (alpha*tmp)>>8;
			pixel = (pixel&(~Amask))|(tmp<<Ashift);
			*outptr++ = pixel;
		}
		inptr += in->pitch/4-inw;
		outptr += out->pitch/4-out->w;
	}

	if (SDL_MUSTLOCK(out)) SDL_UnlockSurface(out);
	if (SDL_MUSTLOCK(in)) SDL_UnlockSurface(in);
	return out;
}

SDL_Surface *Effects::gray(SDL_Surface *in) {
	SDL_Surface *out = SDL_CreateRGBSurface(SDL_SWSURFACE,in->w,in->h,32,in->format->Rmask,in->format->Gmask,in->format->Bmask,in->format->Amask);
	if (SDL_MUSTLOCK(in)) SDL_LockSurface(in);
	if (SDL_MUSTLOCK(out)) SDL_LockSurface(out);

	Uint32 *inptr = (Uint32*)in->pixels;
	Uint32 *outptr = (Uint32*)out->pixels;

	int inw = in->w;
	int inh = in->h;
	int Amask = in->format->Amask;
	int Rshift = in->format->Rshift;
	int Gshift = in->format->Gshift;
	int Bshift = in->format->Bshift;

	for (int y = 0; y < inh; y++) {
		for (int x = 0; x < inw; x++) {
			Uint32 pixel = *inptr++;
			Uint32 alpha = pixel&Amask;
			pixel &= ~Amask;
			Uint8 c1 = pixel>>24;
			Uint8 c2 = (pixel>>16)&0xFF;
			Uint8 c3 = (pixel>>8)&0xFF;
			Uint8 c4 = pixel&0xFF;
			Uint8 val = (c1+c2+c3+c4)/3;
			pixel = alpha | (val<<Rshift) | (val<<Gshift) | (val<<Bshift);
			*outptr++ = pixel;
		}
		inptr += in->pitch/4-inw;
		outptr += out->pitch/4-out->w;
	}

	if (SDL_MUSTLOCK(out)) SDL_UnlockSurface(out);
	if (SDL_MUSTLOCK(in)) SDL_UnlockSurface(in);
	return out;
}


const int linecoords[] = {
	0,33, 18,0, // UPLEFT
	18,0, 57,0, // UP
	57,0, 76,33, // UPRIGHT
	57,65, 76,33, // DOWNRIGHT
	18,65, 57,65, // DOWN
	0,33, 18,65 // DOWNLEFT
};

// float distdata[24]; // a, b, c, sqrt(a*a+b*b)

Uint8 *Effects::highlightbuffer = NULL; // destructor?
float Effects::distdata[24];

float Effects::getDist(int x, int y, int offset) {
	offset *= 4;
	// fabs(a*x+b*y+c)/sqrt(a*a+b*b)
	return fabs(distdata[offset+0]*x+distdata[offset+1]*y+distdata[offset+2])/distdata[offset+3];
}

#define HIGHLIGHTMARGIN 8
#define HIGHLIGHTWIDTH 5
void Effects::precalcHighlight(SDL_Rect *highlightSize) {
	// ax + by + c = 0
	// y-y1 = (y2-y1)/(x2-x1)*(x-x1)
	// (y-y1)*(x2-x1) = (y2-y1)*(x-x1)
	// (y2-y1)*(x-x1) + (y-y1)*(x1-x2) = 0
	// (y2-y1)*x + (x1-x2)*y - (y2-y1)*x1 - (x1-x2)*y1 = 0

	TilePos line = UPLEFT;
	while (line<=DOWNLEFT) {
		int pos = (line-UPLEFT)*4;
		float a = linecoords[pos+3]-linecoords[pos+1]; // y2-y1
		float b = linecoords[pos+0]-linecoords[pos+2]; // x1-x2
		float c = -a*linecoords[pos+0] - b*linecoords[pos+1]; // -a*x1-b*y1
		distdata[pos+0] = a;
		distdata[pos+1] = b;
		distdata[pos+2] = c;
		distdata[pos+3] = sqrt(a*a+b*b);
		line++;
	}

	if (highlightbuffer) delete [] highlightbuffer;
	int outw = 77+2*HIGHLIGHTMARGIN;
	int outh = 66+2*HIGHLIGHTMARGIN;
	highlightbuffer = new Uint8[outw*outh];
	Uint8 *ptr = highlightbuffer;
	if (highlightSize) {
		highlightSize->x = 0;
		highlightSize->y = 0;
		highlightSize->w = outw;
		highlightSize->h = outh;
	}

	for (int y = 0; y < outh; y++) {
		for (int x = 0; x < outw; x++) {

			float angle = 0;
			if ((y-HIGHLIGHTMARGIN)!=33 || (x-HIGHLIGHTMARGIN)!=38)
				angle = 360*atan2(float(33-(y-HIGHLIGHTMARGIN)),x-HIGHLIGHTMARGIN-38)/(2*M_PI);
			if (angle<0) angle += 360;


			if (angle<60)
				line = UPRIGHT;
			else if (angle<120)
				line = UP;
			else if (angle<180)
				line = UPLEFT;
			else if (angle<240)
				line = DOWNLEFT;
			else if (angle<300)
				line = DOWN;
			else
				line = DOWNRIGHT;

			*ptr++ = Uint8(line);
		}
	}
}

SDL_Surface *Effects::createHighlight(const TilePos *which, int maxalpha, int color) {
	SDL_Surface *out = SDL_CreateRGBSurface(SDL_SWSURFACE,77+2*HIGHLIGHTMARGIN,66+2*HIGHLIGHTMARGIN,32,0x00FF0000,0x0000FF00,0x000000FF,0xFF000000);
	if (SDL_MUSTLOCK(out)) SDL_LockSurface(out);

	if (highlightbuffer==NULL) precalcHighlight(NULL);

	Uint32 *outptr = (Uint32*)out->pixels;

	color &= 0x00FFFFFF;

	int outw = out->w;
	int outh = out->h;



	for (int y = 0; y < outh; y++) {
		for (int x = 0; x < outw; x++) {

			TilePos line = TilePos(highlightbuffer[y*outw+x]);

			if (*which!=THIS) {
				const TilePos *ptr = which;
				while (*ptr!=THIS) {
					if (line==*ptr) break;
					ptr++;
				}
				if (*ptr==THIS) {
					*outptr++ = 0;
					continue;
				}
			}

			float dist = getDist(x-HIGHLIGHTMARGIN,y-HIGHLIGHTMARGIN,line-UPLEFT);
			if (dist>HIGHLIGHTWIDTH) *outptr++ = 0;
			else {
				int alpha = int(maxalpha*(HIGHLIGHTWIDTH-dist)/HIGHLIGHTWIDTH);
				*outptr++ = (alpha<<24) | color;
			}
		}
		outptr += out->pitch/4-outw;
	}

	if (SDL_MUSTLOCK(out)) SDL_UnlockSurface(out);
	return out;
}

SDL_Surface *Effects::createHighlight(TilePos which, int maxalpha, int color) {
	TilePos array[2];
	array[0] = which;
	array[1] = THIS;
	return createHighlight(array,maxalpha,color);
}


static inline void fetch(Uint8 *inptr, SDL_PixelFormat *fmt, Uint8 &a, Uint8 &r, Uint8 &g, Uint8 &b) {
#ifdef ROTATEARGB
	Uint32 pixel = *(Uint32*)inptr;
	a = pixel>>24;
	r = (pixel>>16)&0xFF;
	g = (pixel>>8)&0xFF;
	b = pixel&0xFF;
#else
	Uint32 pixel = *(Uint32*)inptr;
#if SDL_BYTEORDER==SDL_BIG_ENDIAN
	pixel >>= 4-fmt->BytesPerPixel;
#endif
	if (fmt->Amask==0)
		a = 0xFF;
	else
		a = (pixel&fmt->Amask)>>fmt->Ashift;
	r = (pixel&fmt->Rmask)>>fmt->Rshift;
	g = (pixel&fmt->Gmask)>>fmt->Gshift;
	b = (pixel&fmt->Bmask)>>fmt->Bshift;
#endif
}

static Uint32 blend(Uint8 *inptr, SDL_Surface *in, int posX, int posY) {
	Uint8 a1, r1, g1, b1;
	Uint8 a2, r2, g2, b2;
	Uint8 a3, r3, g3, b3;
	Uint8 a4, r4, g4, b4;
	int x = posX>>16;
	int y = posY>>16;
	inptr += y*in->pitch + x*in->format->BytesPerPixel;
	int bpp = in->format->BytesPerPixel;
	int pitch = in->pitch;
	if (x<0) {
		a1 = a3 = 0;
		if (x+1==0) {
			if (y<0) {
				a2 = 0;
				if (y+1==0) {
					fetch(inptr+pitch+bpp,in->format,a4,r4,g4,b4);
				} else {
					a4 = 0;
				}
			} else if (y>=in->h) {
				a2 = a4 = 0;
			} else {
				fetch(inptr+bpp,in->format,a2,r2,g2,b2);
				if (y+1<in->h) {
					fetch(inptr+pitch+bpp,in->format,a4,r4,g4,b4);
				} else {
					a4 = 0;
				}
			}
		} else {
			a2 = a4 = 0;
		}
	} else if (x>=in->w) {
		a1 = a2 = a3 = a4 = 0;
	} else {
		if (y<0) {
			a1 = a2 = 0;
			if (y+1==0) {
				fetch(inptr+pitch,in->format,a3,r3,g3,b3);
				if (x+1<in->w) {
					fetch(inptr+pitch+bpp,in->format,a4,r4,g4,b4);
				} else {
					a4 = 0;
				}
			} else {
				a3 = a4 = 0;
			}
		} else if (y>=in->h) {
			a1 = a2 = a3 = a4 = 0;
		} else {
			fetch(inptr,in->format,a1,r1,g1,b1);
			if (x+1<in->w) {
				fetch(inptr+bpp,in->format,a2,r2,g2,b2);
				if (y+1<in->h) {
					fetch(inptr+pitch,in->format,a3,r3,g3,b3);
					fetch(inptr+pitch+bpp,in->format,a4,r4,g4,b4);
				} else {
					a3 = a4 = 0;
				}
			} else {
				a2 = a4 = 0;
				if (y+1<in->h) {
					fetch(inptr+pitch,in->format,a3,r3,g3,b3);
				} else {
					a3 = 0;
				}
			}
		}
	}

	int xFrac = (posX>>8)&0xFF;
	int yFrac = (posY>>8)&0xFF;

	int fact1 = a1*(0x100-xFrac)*(0x100-yFrac);
	int fact2 = a2*xFrac*(0x100-yFrac);
	int fact3 = a3*(0x100-xFrac)*yFrac;
	int fact4 = a4*xFrac*yFrac;

	Uint8 a = (fact1+fact2+fact3+fact4)>>16;
	Uint16 r, g, b;

	if (a!=0) {
		r = (r1*fact1+r2*fact2+r3*fact3+r4*fact4)>>16;
		g = (g1*fact1+g2*fact2+g3*fact3+g4*fact4)>>16;
		b = (b1*fact1+b2*fact2+b3*fact3+b4*fact4)>>16;
		r /= a;
		g /= a;
		b /= a;
	} else {
		r = g = b = 0;
	}



	return (a<<24) | (r<<16) | (g<<8) | b;
}

SDL_Surface *Effects::rotate(SDL_Surface *in, double angle, bool smooth) {
	double cosa = cos(angle);
	double sina = sin(angle);
	int diag1W = abs((int)(in->w*cosa-in->h*sina));
	int diag1H = abs((int)(in->h*cosa+in->w*sina));
	int diag2W = abs((int)((-in->w)*cosa-in->h*sina));
	int diag2H = abs((int)(in->h*cosa+(-in->w)*sina));
	if (diag2W>diag1W) diag1W = diag2W;
	if (diag2H>diag1H) diag1H = diag2H;
	SDL_Surface *out = SDL_CreateRGBSurface(SDL_SWSURFACE|SDL_RLEACCEL,diag1W,diag1H,32,0x00FF0000,0x0000FF00,0x000000FF,0xFF000000);


	int corner1X = (int)((1<<16)*(in->w + (-out->w)*cosa + (-out->h)*sina)/2);
	int corner1Y = (int)((1<<16)*(in->h + (-out->h)*cosa - (-out->w)*sina)/2);
	int corner2X = (int)((1<<16)*(in->w +  (out->w)*cosa + (-out->h)*sina)/2);
	int corner2Y = (int)((1<<16)*(in->h + (-out->h)*cosa -  (out->w)*sina)/2);
	int corner3X = (int)((1<<16)*(in->w + (-out->w)*cosa +  (out->h)*sina)/2);
	int corner3Y = (int)((1<<16)*(in->h +  (out->h)*cosa - (-out->w)*sina)/2);

	int XdiffX = (corner2X-corner1X)/out->w;
	int XdiffY = (corner2Y-corner1Y)/out->w;
	int YdiffX = (corner3X-corner1X)/out->h;
	int YdiffY = (corner3Y-corner1Y)/out->h;

	if (SDL_MUSTLOCK(in)) SDL_LockSurface(in);
	if (SDL_MUSTLOCK(out)) SDL_LockSurface(out);

	Uint32 *outptr = (Uint32*) out->pixels;
	Uint8 *inptr = (Uint8*) in->pixels;
	int pitch = in->pitch;
#ifndef ROTATEARGB
	int bpp = in->format->BytesPerPixel;
	Uint8 a, r, g, b;
	a = r = g = b = 0;
#endif
	int inw = in->w;
	int inh = in->h;

	if (smooth) {
		for (int y = 0; y < out->h; y++) {
			int posX = corner1X + y*YdiffX;
			int posY = corner1Y + y*YdiffY;
			for (int x = 0; x < out->w; x++) {
				*outptr++ = blend(inptr,in,posX,posY);
				posX += XdiffX;
				posY += XdiffY;
			}
			outptr += out->pitch/4-out->w;
		}
	} else {
		for (int y = 0; y < out->h; y++) {
			int posX = corner1X + y*YdiffX;
			int posY = corner1Y + y*YdiffY;
			for (int x = 0; x < out->w; x++) {
				int x2 = posX>>16;
				int y2 = posY>>16;
#ifdef ROTATEARGB
				if (x2<0 || x2>=inw || y2<0 || y2>=inh)
					*outptr++ = 0;
				else
					*outptr++ = *((Uint32*)(inptr + y2*pitch + x2*4));
#else
				if (x2<0 || x2>=inw || y2<0 || y2>=inh) {
					a = 0;
				} else {
					Uint32 pixel = *((Uint32*)(inptr + y2*pitch + x2*bpp));
#if SDL_BYTEORDER==SDL_BIG_ENDIAN
					pixel >>= 4-bpp;
#endif
					if (in->format->Amask==0)
						a = 0xFF;
					else
						a = (pixel&in->format->Amask)>>in->format->Ashift;
					r = (pixel&in->format->Rmask)>>in->format->Rshift;
					g = (pixel&in->format->Gmask)>>in->format->Gshift;
					b = (pixel&in->format->Bmask)>>in->format->Bshift;
				}
				*outptr++ = (a<<24) | (r<<16) | (g<<8) | b;
#endif
				posX += XdiffX;
				posY += XdiffY;
			}
			outptr += out->pitch/4-out->w;
		}
	}

	if (SDL_MUSTLOCK(out)) SDL_UnlockSurface(out);
	if (SDL_MUSTLOCK(in)) SDL_UnlockSurface(in);
	return out;
}



